local _G = getfenv(0)

ElkBuffBars = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceDB-2.0", "AceEvent-2.0")
ElkBuffBars:RegisterDB("ElkBuffBarsDB")

local updates_disabled
local scan_happend

function ElkBuffBars:OnInitialize()
	local build = select(2, _G.GetBuildInfo())
	if not self.db.account.build or self.db.account.build ~= build then
		self.db.account.build = build
		self.db.account.maxtimes = {
			BUFF = {},
			DEBUFF = {},
			TENCH = {},
		}
	end
end

function ElkBuffBars:OnEnable()
	-- create bargroups based on stored settings
	self:CreateBarGroups()
	
	-- register events
	self:RegisterEvent("PLAYER_AURAS_CHANGED")
	self:ScheduleRepeatingEvent(self.DoUpdates, .1, self)
	
	--
	self:ScanData_BUFF()
	self:ScanData_DEBUFF()
	self:ScanData_TENCH()
	self:ScanData_TRACKING()
end

function ElkBuffBars:OnDisable()
	-- recycle all bargroups
	self:RemoveBarGroups()
end


--
--
--
local barcache = {}

function ElkBuffBars:GetBar()
	if #barcache > 0 then
		return table.remove(barcache, #barcache)
	else
		return EBB_Bar:new()
	end
end

function ElkBuffBars:RecycleBar(bar)
	bar:Reset()
	table.insert(barcache, bar)
end

--
--
--
local bargroupcache = {}

function ElkBuffBars:GetBarGroup()
	if #bargroupcache > 0 then
		return table.remove(bargroupcache, #bargroupcache)
	else
		return EBB_BarGroup:new()
	end
end

function ElkBuffBars:RecycleBarGroup(bargroup)
	bargroup:Reset()
	table.insert(bargroupcache, bargroup)
end

--
--
--
local datatablecache = {}

local function GetDataTable()
	if #datatablecache > 0 then
		return table.remove(datatablecache, #datatablecache)
	else
		return {}
	end
end

local function RecycleDataTable(dt)
	table.insert(datatablecache, dt)
end

--
--
--
function ElkBuffBars:CreateBarGroups()
	if not self.bargroups then
		self.bargroups = {}
	else
		for k, v in pairs(self.bargroups) do
			self:RecycleBarGroup(v)
			self.bargroups[k] = nil
		end
	end

-- -----
	self.db.profile.bargroups = {}
	local layout = {
		["bars"] = {
			["icon"]				= "LEFT",											-- "LEFT", "RIGHT", false
			["iconcount"]			= true,												-- true, false
			["iconcountanchor"]		= "BOTTOMRIGHT",									-- <anchor>
			["iconcountfont"]		= "Fonts\\ARIALN.TTF",								-- <font>
			["iconcountfontsize"]	= 14,												-- <font size>
			["bar"]					= "Interface\\Addons\\ElkBuffBar\\BarTexture",		-- <texture path>, false
			["textTL"]				= "NAMERANKCOUNT",									-- false, "NAME", "NAMERANK", "NAMECOUNT", "NAMERANKCOUNT", "RANK", "COUNT", "TIMELEFT", "DEBUFFTYPE"
			["textTLfont"]			= "Fonts\\FRIZQT__.TTF",							-- <font>
			["textTLfontsize"]		= 14,												-- <font size>
			["textTLcolor"]			= {1, 1, 1},										-- <color set>
			["textTLalign"]			= "LEFT",											-- left, center, right
			["textTR"]				= "TIMELEFT",										-- false, "NAME", "NAMERANK", "NAMECOUNT", "NAMERANKCOUNT", "RANK", "COUNT", "TIMELEFT", "DEBUFFTYPE"
			["textTRfont"]			= "Fonts\\FRIZQT__.TTF",							-- <font>
			["textTRfontsize"]		= 14,												-- <font size>
			["textTRcolor"]			= {1, 1, 1},										-- <color set>
			["barcolor"]			= {0.3, 0.5, 1},									-- <color set>
			["barbgcolor"]			= {0, 0.5, 1},										-- <color set>
			["debufftypecolor"]		= true,												-- true, false
			["timeformat"]			= "CONDENSED",										-- "DEFAULT", "EXTENDED", "FULL", "SHORT", "CONDENSED"
			["width"]				= 350,
			["height"]				= 25,
		},
		["filter"] = {
			["type"] = {
				["BUFF"] = true,
				["TENCH"] = true,
				["TRACKING"] = true,
			}
		},
		growup = false,
		barspacing = 0,
		anchortext = "BUFFS & Co.",
	}
	table.insert(self.db.profile.bargroups, layout)
	layout = {
		["bars"] = {
			["icon"]				= "LEFT",											-- "LEFT", "RIGHT", false
			["iconcount"]			= true,												-- true, false
			["iconcountanchor"]		= "BOTTOMRIGHT",									-- <anchor>
			["iconcountfont"]		= "Fonts\\ARIALN.TTF",								-- <font>
			["iconcountfontsize"]	= 14,												-- <font size>
			["bar"]					= "Interface\\AddOns\\Chronometer\\Textures\\Charcoal.tga",		-- <texture path>, false
			["textTL"]				= "NAMERANK",										-- false, "NAME", "NAMERANK", "NAMECOUNT", "NAMERANKCOUNT", "RANK", "COUNT", "TIMELEFT", "DEBUFFTYPE"
			["textTLfont"]			= "Fonts\\MORPHEUS.TTF",							-- <font>
			["textTLfontsize"]		= 20,												-- <font size>
			["textTLcolor"]			= {1, 1, 1},										-- <color set>
			["textTLalign"]			= "LEFT",											-- left, center, right
			["textTR"]				= "COUNT",											-- false, "NAME", "NAMERANK", "NAMECOUNT", "NAMERANKCOUNT", "RANK", "COUNT", "TIMELEFT", "DEBUFFTYPE"
			["textTRfont"]			= "Fonts\\FRIZQT__.TTF",							-- <font>
			["textTRfontsize"]		= 14,												-- <font size>
			["textTRcolor"]			= {1, 1, 1},										-- <color set>
			["textBL"]				= "DEBUFFTYPE",										-- false, "NAME", "NAMERANK", "NAMECOUNT", "NAMERANKCOUNT", "RANK", "COUNT", "TIMELEFT", "DEBUFFTYPE"
			["textBLfont"]			= "Fonts\\FRIZQT__.TTF",							-- <font>
			["textBLfontsize"]		= 14,												-- <font size>
			["textBLcolor"]			= {1, 1, 1},										-- <color set>
			["textBLalign"]			= "LEFT",											-- left, center, right
			["textBR"]				= "TIMELEFT",										-- false, "NAME", "NAMERANK", "NAMECOUNT", "NAMERANKCOUNT", "RANK", "COUNT", "TIMELEFT", "DEBUFFTYPE"
			["textBRfont"]			= "Fonts\\FRIZQT__.TTF",							-- <font>
			["textBRfontsize"]		= 14,												-- <font size>
			["textBRcolor"]			= {1, 1, 1},										-- <color set>
			["barcolor"]			= {1, 0, 0},									-- <color set>
			["barbgcolor"]			= {1, 0, 0},										-- <color set>
			["debufftypecolor"]		= true,												-- true, false
			["timeformat"]			= "CONDENSED",										-- "DEFAULT", "EXTENDED", "FULL", "SHORT", "CONDENSED"
			["width"]				= 350,
			["height"]				= 50,
		},
		["filter"] = {
			["type"] = {
				["DEBUFF"] = true,
			}
		},
		growup = false,
		barspacing = 0,
		anchortext = "DEBUFFS",
		stickto = 1
	}
	table.insert(self.db.profile.bargroups, layout)
-- -----
	local bglayouts = self.db.profile.bargroups
	for k, v in ipairs(bglayouts) do
		table.insert(self.bargroups, self:GetBarGroup())
	end
	for k, v in ipairs(self.bargroups) do
		local layout = bglayouts[k]
		v:SetLayout(layout)
		local container = v:GetContainer()
		container:Show()
		if layout.stickto then
			container:SetPoint(layout.growup and "BOTTOMLEFT" or "TOPLEFT", self.bargroups[layout.stickto]:GetContainer(), layout.growup and "TOPLEFT" or "BOTTOMLEFT")
		elseif layout.x and layout.y then
			container:SetPoint(layout.growup and "BOTTOMLEFT" or "TOPLEFT", UIParent, "BOTTOMLEFT", layout.x, layout.y)
		else
			container:SetPoint("CENTER", UIParent, "CENTER", layout.x, layout.y)
			v:ToggleAnchor(true)
		end
	end
end

function ElkBuffBars:RemoveBarGroups()
	for k, v in self.bargroups do
		self:RecycleBarGroup(v)
		self.bargroups[k] = nil
	end
end

--
--
--
ElkBuffBars.buffdata = {}
ElkBuffBars.debuffdata = {}
ElkBuffBars.tenchdata = {}
ElkBuffBars.trackingdata = {}

function ElkBuffBars:PLAYER_AURAS_CHANGED()
	self:ScanData_BUFF()
	self:ScanData_DEBUFF()
	self:ScanData_TRACKING()
end

function ElkBuffBars:ScanData_BUFF()
	local maxtimes = self.db.account.maxtimes.BUFF
	for k, v in pairs(self.buffdata) do
		RecycleDataTable(v)
		self.buffdata[k] = nil
	end
	local i = 1
	while true do
		local buffIndex, untilCancelled = GetPlayerBuff(i, "HELPFUL")
		if buffIndex == 0 then break end
		local timeleft = GetPlayerBuffTimeLeft(buffIndex)
		local timemax = timeleft
		local name, rank = GetPlayerBuffName(buffIndex)
		local tname = name..(rank and (" "..rank) or "")
		if rank then
			_, _, rank = string.find(rank, RANK.." (%d+)")
		end
		if name then
			if maxtimes[tname] and timemax < maxtimes[tname] then
				timemax = maxtimes[tname]
			else
				maxtimes[tname] = timemax
			end
		end

		local dt = GetDataTable()
		dt.id				= buffIndex
		dt.name				= name
		dt.realname			= name
		dt.rank				= rank and tonumber(rank) or nil
		dt.type				= "BUFF"
		dt.realtype			= "BUFF"
		dt.debufftype		= nil
		dt.timeleft			= timeleft
		dt.timemax			= timemax
		dt.untilcancelled	= (untilCancelled == 1) and true or nil
		dt.charges			= GetPlayerBuffApplications(buffIndex)
		dt.icon				= GetPlayerBuffTexture(buffIndex)
		
		table.insert(self.buffdata, dt)
		i = i + 1
	end
	scan_happend = true
end

function ElkBuffBars:ScanData_DEBUFF()
	local maxtimes = self.db.account.maxtimes.DEBUFF
	for k, v in pairs(self.debuffdata) do
		RecycleDataTable(v)
		self.debuffdata[k] = nil
	end
	local i = 1
	while true do
		local buffIndex, untilCancelled = GetPlayerBuff(i, "HARMFUL")
		if buffIndex == 0 then break end
		local timeleft = GetPlayerBuffTimeLeft(buffIndex)
		local timemax = timeleft
		local name, rank = GetPlayerBuffName(buffIndex)
		local tname = name..(rank and (" "..rank) or "")
		if rank then
			_, _, rank = string.find(rank, RANK.." (%d+)")
		end
		if name then
			if maxtimes[tname] and timemax < maxtimes[tname] then
				timemax = maxtimes[tname]
			else
				maxtimes[tname] = timemax
			end
		end

		local dt = GetDataTable()
		dt.id				= buffIndex
		dt.name				= name
		dt.realname			= name
		dt.rank				= rank and tonumber(rank) or nil
		dt.type				= "DEBUFF"
		dt.realtype			= "DEBUFF"
		dt.debufftype		= GetPlayerBuffDispelType(buffIndex)
		dt.timeleft			= timeleft
		dt.timemax			= timemax
		dt.untilcancelled	= (untilCancelled == 1) and true or nil
		dt.charges			= GetPlayerBuffApplications(buffIndex)
		dt.icon				= GetPlayerBuffTexture(buffIndex)
		
		table.insert(self.debuffdata, dt)
		i = i + 1
	end
	scan_happend = true
end

function ElkBuffBars:ScanData_TENCH()
	local maxtimes = self.db.account.maxtimes.TENCH
	for k, v in pairs(self.tenchdata) do
		RecycleDataTable(v)
		self.tenchdata[k] = nil
	end
	local hasMainHandEnchant, mainHandExpiration, mainHandCharges, hasOffHandEnchant, offHandExpiration, offHandCharges = GetWeaponEnchantInfo()
	if ( hasMainHandEnchant ) then
		local timeleft = mainHandExpiration / 1000
		local timemax = timeleft
	
		local id = GetInventorySlotInfo("MainHandSlot")
		local name, rank = self:GetTempBuffName(id)
		local tname = name..(rank and (" "..rank) or "")
		if rank then
			_, _, rank = string.find(rank, RANK.." (%d+)")
		end
		if name then
			if maxtimes[tname] and timemax < maxtimes[tname] then
				timemax = maxtimes[tname]
			else
				maxtimes[tname] = timemax
			end
		end
		local dt = GetDataTable()
		dt.id				= id
		dt.name				= name
		dt.realname			= name
		dt.rank				= rank and tonumber(rank) or nil
		dt.type				= "TENCH"
		dt.realtype			= "TENCH"
		dt.debufftype		= nil
		dt.timeleft			= timeleft
		dt.timemax			= timemax
		dt.untilcancelled	= nil
		dt.charges			= mainHandCharges
		dt.icon				= GetInventoryItemTexture("player", id)
		
		table.insert(self.tenchdata, dt)
	end
	if ( hasOffHandEnchant ) then
		local timeleft = offHandExpiration / 1000
		local timemax = timeleft

		local id = GetInventorySlotInfo("SecondaryHandSlot")
		local name, rank = self:GetTempBuffName(id)
		local tname = name..(rank and (" "..rank) or "")
		if rank then
			_, _, rank = string.find(rank, RANK.." (%d+)")
		end
		if name then
			if maxtimes[tname] and timemax < maxtimes[tname] then
				timemax = maxtimes[tname]
			else
				maxtimes[tname] = timemax
			end
		end
		local dt = GetDataTable()
		dt.id				= id
		dt.name				= name
		dt.realname			= name
		dt.rank				= rank and tonumber(rank) or nil
		dt.type				= "TENCH"
		dt.realtype			= "TENCH"
		dt.debufftype		= nil
		dt.timeleft			= timeleft
		dt.timemax			= timemax
		dt.untilcancelled	= nil
		dt.charges			= mainHandCharges
		dt.icon				= GetInventoryItemTexture("player", id)
		
		table.insert(self.tenchdata, dt)
	end
	scan_happend = true
end

function ElkBuffBars:ScanData_TRACKING()
	for k, v in pairs(self.trackingdata) do
		RecycleDataTable(v)
		self.trackingdata[k] = nil
	end
	local icon = GetTrackingTexture()
	if icon then
		name = self:GetTrackingName()
		local dt = GetDataTable()
		dt.id				= id
		dt.name				= name
		dt.realname			= name
		dt.rank				= nil
		dt.type				= "TRACKING"
		dt.realtype			= "TRACKING"
		dt.debufftype		= nil
		dt.timeleft			= 1
		dt.timemax			= 1
		dt.untilcancelled	= true
		dt.charges			= 1
		dt.icon				= icon
		
		table.insert(self.trackingdata, dt)
	end
	scan_happend = true
end

function ElkBuffBars:GetTempBuffName(id)
	return "", ""
end

function ElkBuffBars:GetTrackingName()
	return ""
end

--
--
--
function ElkBuffBars:DoUpdates()
	if updates_disabled then return end
	
	for k, v in pairs(self.buffdata) do
		if not v.untilcancelled then
			local timeleft = GetPlayerBuffTimeLeft(v.id)
			if timeleft > v.id then
				self:ScanData_BUFF()
				break
			end
			v.timeleft = timeleft
		end
	end

	for k, v in pairs(self.debuffdata) do
		if not v.untilcancelled then
			local timeleft = GetPlayerBuffTimeLeft(v.id)
			if timeleft > v.id then
				self:ScanData_DEBUFF()
				break
			end
			v.timeleft = timeleft
		end
	end

	-- check for changed weaponbuffs
	
	-- if we did a scan, update bar data
	if scan_happend then
		for k, v in pairs(self.bargroups) do
			v:UpdateData()
		end
		scan_happend = nil
	else
		for k, v in pairs(self.bargroups) do
			v:UpdateTimeleft()
		end
	end
end


function ElkBuffBars:StickGroup(bargroup)
end
